/* hexsplit.c			letzte nderung: 15.9.2012 */
static char *VERSION="0.0";
/*
Aufteilen einer grossen hex-Datei in mehrere kleinere.
Programmiert nach den Angaben von dieser Webseite:
 http://de.wikipedia.org/wiki/Intel_HEX

History:
Erstellung:	15.9.2012 (RPf)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#define MAXARG 1

/************************ Vordeklarationen *******************************/
void hexsplit(FILE *fp,const char *name1);
/*************************************************************************/

static char argflag['Z'+1];

void usage()
{
 printf("hexsplit Version %s\n",VERSION);
 printf("Anwendung: hexsplit name.hex [-Flags]\n");
 printf("  Flags:\n");
 printf("      h = Hilfe\n");
 printf("      a = Anpassung fuer fehlerhaftes avrdude-5.10\n");
 if(argflag['H'])
  {
   printf("Die Datei name.hex wird aufgeteilt in name_a.hex, name_b.hex ...\n");
   printf("Jede Ausgabedatei representiert dann maximal 64KB Daten.\n");
  }
 exit(0);
}

void setflags(char *s)
{
 int c;
 while((c= *s++)!=0)
  if((c=toupper(c))<='Z')  argflag[c]=1;
}

const char *rstring="rb",*wstring="wb"; //je nach System "rb" und "wb" oder dann "r" und "w"

int main(int argc,char *argv[])
{
 FILE *fp;
 int i,j;
 char name1[80];
 for(name1[0]=0,i=j=1;i<argc;i++)
   if(*argv[i]=='-') setflags(argv[i]);
   else strcpy(name1,argv[i]);
 if(j>MAXARG+1 || name1[0]=='?' || argflag['?'] || argflag['H']) usage();
 if(name1[0]==0) {printf("Hexdatei:"); i=scanf("%s",name1);}
 if(!(fp=fopen(name1,rstring)))
	{printf("'%s' nicht gefunden\n",name1); exit(0);}
 hexsplit(fp,name1);
 fclose(fp);
 return 0;
}

/*************************** kleinkram ***********************************/
int ascitodigit(int c)
{		/* wandelt ein ASCI-Zeichen in eine 1-stellige Hex-Zahl */
 if(!isxdigit(c)) return 0;
 if(islower(c)) c=toupper(c)-'0';
 else c-='0';
 if(c>9) c += '0'-'A'+10;
 return c;
}

#define getline mygetline

int getline(FILE *fp,char *s,int lim)
{		/* liest eine Textzeile oder maximal lim Zeichen */
		/* und ersetzt den Zeilentrenner durch 0         */
 int c=EOF;
 while(--lim && (c=getc(fp))!=EOF && c!='\n')
    {if(c!='\r')  *s++ = c;}
 *s='\0';
 return (c!=EOF);	/* TRUE wenn erfolgreich, FALSE wenn Fileende */
}

void putline(FILE *fp,const char *s) /* wie fprintf(fp,"%s\n",s)  aber */
{				     /* ohne ein Nullbyte zu schreiben */
 char c;
 while((c= *s++)!=0)  putc(c,fp);
 putc('\n',fp);
}

/*************************** Hauptteil ***********************************/
int pruefsummecheck(const char *s)
{
 int pruefsumme=0,c1,c2;
 while((c1= *s++)!=0 && (c2= *s++)!=0)
  {
   pruefsumme += (ascitodigit(c1)<<4)+ascitodigit(c2);
  }
 return (pruefsumme&0xFF);
}

void hexsplit(FILE *fp,const char *name1)
{
 int c,i,j,reclen,rectyp;
 long adresse= -1, extadresse=0, adr;
 char zeile[120],name2[80];
 char zeile1neu[120];
 char buchstabe='a';
 FILE *fp2=NULL;
 zeile1neu[0]=0; //erste Zeile fuer neue Datei
 while(getline(fp,zeile,120))
  {
   if(zeile[0]!=':') {printf("Fehler: unbekannte Zeile '%s'\n",zeile); continue;}
   if(pruefsummecheck(&zeile[1])!=0) {printf(" Fehler: Pruefsumme falsch in Zeile '%s'\n",zeile); return;}
   j=1;
   reclen = (ascitodigit(zeile[j])<<4) + ascitodigit(zeile[j+1]);
   j+=2;
   adr = (ascitodigit(zeile[j])<<12)+(ascitodigit(zeile[j+1])<<8)+(ascitodigit(zeile[j+2])<<4)+ascitodigit(zeile[j+3]);
   j+=4;
   rectyp = (ascitodigit(zeile[j])<<4) + ascitodigit(zeile[j+1]);
   j+=2;
   if(fp2==NULL) //wenn Ausgabedatei noch nicht offen
    {
     int k;
     for(i=k=0;(c=name1[i])!=0;i++) //neuer Name erzeugen
      {
       if(c=='.') {name2[k++]='_'; name2[k++]=buchstabe++;} //vor .hex ein _a einfuegen (naechstes mal _b ...)
       name2[k++]=c;
      }
     if(!(fp2=fopen(name2,wstring)))
        {printf("kann '%s' nicht erstellen\n",name2); return;}
     if(zeile1neu[0]!=0) //wenn erste Zeile fuer neue Datei nicht leer ist
        {putline(fp2,zeile1neu); zeile1neu[0]=0;} //dann in die neue Datei speichern
    }
   if(rectyp==0) //Data Record, Nutzdaten
    {
     if(adresse<0) adresse=adr;
     else if(adr>adresse)
      {
       printf("Warnung: Luecke im Datensatz zwischen %ld und %ld\n",adresse,adr);
       adresse=adr;
      }
     else if(adr<adresse)
      {
       printf("Fehler: Adresse neu gesetzt von %ld auf %ld\n",adresse,adr);
       adresse=adr;
      }
     for(i=0;i<reclen;i++)
      {
       c = (ascitodigit(zeile[j])<<4) + ascitodigit(zeile[j+1]);
       j+=2;
       if(++adresse==0x10000) {adresse=0; extadresse+=0x10000;}
      }
     putline(fp2,zeile);
    }
   else if(rectyp==1) //End of File
    {
     putline(fp2,zeile);
     fclose(fp2); fp2=NULL;
     printf("End of File Record - Umwandlung abgeschlossen.\n");
     return;
    }
   else if(rectyp==2) //Extended Segment Address Record
    {
     extadresse = (ascitodigit(zeile[j])<<12)+(ascitodigit(zeile[j+1])<<8)+(ascitodigit(zeile[j+2])<<4)+ascitodigit(zeile[j+3]);
     extadresse *= 16;
     printf("Test: extadresse = %ld = 0x%lX\n",extadresse,extadresse); //test
     putline(fp2,":00000001FF"); //Ein "End of File Record" anfuegen
     fclose(fp2); fp2=NULL; //aktuelle Ausgabedatei schliessen
     if(argflag['A'])
      {
       int neuepruefsumme;
       extadresse = 0x40000-extadresse; //Anpassung fuer fehlerhaftes avrdude-5.10
       sprintf(zeile,":02000002%04lX",extadresse/16);
       neuepruefsumme = -pruefsummecheck(&zeile[1]);
       sprintf(zeile,":02000002%04lX%02X",extadresse/16,neuepruefsumme&0xFF);
      }
     strcpy(zeile1neu,zeile); //aktuelle Zeile zur ersten Zeile fuer neue Datei kopieren
    }
   else if(rectyp==3) //Start Segment Address Record
    {
     putline(fp2,zeile);
     printf("Start Segment Address Record - noch nicht unterstuetzt. Unveraendert kopiert.\n");
    }
   else if(rectyp==4) //Extended Linear Address Record
    {
     putline(fp2,zeile);
     printf("Extended Linear Address Record - noch nicht unterstuetzt. Unveraendert kopiert.\n");
    }
   else if(rectyp==5) //Start Linear Address Record
    {
     putline(fp2,zeile);
     printf("Start Linear Address Record - noch nicht unterstuetzt. Unveraendert kopiert.\n");
    }
  }
 if(fp2!=NULL)
  {
   printf("Fehler: kein \"End of File Record\" gefunden.\n");
   fclose(fp2);
  }
}
